unit butler;

{
  Usermode driver for the Topcom Butler 4012/4812 usb voip phone.
  
  Copyright (C) 2007 Luca Olivetti <luca@ventoso.org>

  This source is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your option)
  any later version.

  This code is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
  details.

  A copy of the GNU General Public License is available on the World Wide Web
  at <http://www.gnu.org/copyleft/gpl.html>. You can also obtain it by writing
  to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
  MA 02111-1307, USA.
}

{

This class uses thead an synchronize. For it to work the procedure
CheckSynchronize has to be called periodically in the main loop.
If you're writing a GUI application with Lazarus it'll be done automatically
but if you're writing a console application you need to call it yourself.

The usage is very simple: just create an onbject of type TButlerPhone and
assign event handlers if you want to be informed of events. Then, to begin
communication with the phone, call the method 'Start'.

I've only tested with my phone (a 4012 with only the voip functionality, no
landline connector) but it should work also with the 4812.

I couldn't find a way to activate the voicemail icon.

}

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, libusb, syncobjs, simplelogger;

type


  // The status is displayed when the phone is idle in the selected
  // language
  TButlerStatus = (BSOffline, //keypresses aren't reported to the pc in this state
                   BSOnline, BSSkypeMe, BSAway, BSNotAvailable,
                   BSDoNotDisturb, BSInvisible, BSLogOff,
                   //most features of the phone don't work in the following
                   //states (e.g. caller id display, calls status icon)
                   BSClearDisplay:=16,
                   BSMSN:=17,
                   BSSJPhone:=33,
                   BSNet2Phone:=49);

  // Calls icon (only works if in state BSOffline..BSLogOff)
  TCallsStatus = (CSNoCalls,   // icon off
                  CSCalls,     // icon on
                  CSNewCalls); // icon blinking

  //Key for the "calls", "redial" and "contacts" events
  //CKExit is only fired by the "calls" key
  TContactKey = (CKFirst, CKUp, CKDown, CKExit);
  
  
  TButlerPhone = class;
  

  { TConnectThread
    Helper thread to wait for the connection of the phone.
  }
  
  TConnectThread = class(TThread)
  private
    FUsbDevHandle: Pusb_dev_handle;
    FOwner: TButlerPhone;
    procedure SyncProc;
    procedure Execute;override;
    function FindDevice:pusbdevice;
    function InitDevice:boolean;
  public  
    constructor create(AOwner:TButlerPhone);
  end;
  
  { TSendFifo
    Fifo accessible from different threads. It signals the
    send thread when the first message is available.
   }

  TSendFifo = class
  private
    FAccess: TCriticalSection;
    FFifo: TStringList;
    FWaiter: TEventObject;
  public  
    procedure Push(s:string);
    function Pop(var s:string):boolean;
    function IsEmpty:boolean;
    constructor Create;
    destructor Destroy;override;
    property Waiter:TEventObject write FWaiter;
  end;

  { TSendThread
    this thread takes messages from the send fifo and
    sends them to the phone.
  }

  TSendThread = class(TThread)
  private
    FUsbDevHandle: Pusb_dev_handle;
    FSendFifo: TSendFifo;
    FWaiter: TEventObject;
    FTerminated:boolean;
    procedure execute;override;
  public  
    constructor create(UsbDevHandle:Pusb_dev_handle; SendFifo:TSendFifo);
    procedure Terminate;
  end;
  

  { TStatusThread
    This thread waits for a message from the phone.
    In case of error it decides that the phone has been disconnected.
   }

  TStatusThread = class(TThread)
  private
    FOwner: TButlerPhone;
    FUsbDevHandle:Pusb_dev_handle;
    FData:string;
    procedure Receive;
    procedure Error;
    procedure execute;override;
  public  
    constructor create(AOwner:TButlerPhone;UsbDevHandle:Pusb_dev_handle);
  end;
  
  { TRingerThread
    The ring command has to be sent at regular intervals, this thread
    takes care of it.
   }

  TRingerThread = class(TThread)
  private
    FOwner: TButlerPhone;
    procedure SyncProc;
    procedure execute;override;
  public  
    constructor create(AOwner:TButlerPhone);
  end;
  
  {
   Specific event handlers for the phone
  }

  TNotifyDial = procedure(Sender:TObject; DialString:string) of object;
  TNotifyKey = procedure(Sender:TObject; Key:TContactKey) of object;

  { TButlerPhone }

  TButlerPhone = class
  private
    FUsbDevHandle: Pusb_dev_handle;
    FButlerStatus : TButlerStatus;
    FCallsStatus: TCallsStatus;
    FSendFifo: TSendFifo;
    FSendThread: TSendThread;
    FStatusThread: TStatusThread;
    FRingerThread: TRingerThread;
    FConnectThread:TConnectThread;
    FLastCommand:char;
    FConnected:boolean;
    FOnConnected: TNotifyEvent;
    FOnDisconnected: TNotifyEvent;
    FOnDialNumber: TNotifyDial;
    FOnDialName: TNotifyDial;
    FOnAcceptCall: TNotifyEvent;
    FOnHangup: TNotifyEvent;
    FOnDigit: TNotifyDial;
    FOnContactsKey: TNotifyKey;
    FOnCallsKey: TNotifyKey;
    FOnRedialKey: TNotifyKey;
    FOnShowIdKey: TNotifyDial;
    FIgnoreChecksumErrors: boolean;
    procedure Send(s:string);
    procedure SendStatus;
    procedure SetButlerStatus(newValue:TButlerStatus);
    procedure SetCallsStatus(newValue:TCallsStatus);
    function DecodeNumber(digits:integer; code:string):string;
    procedure Command(s:string);
    procedure receive(s:string);
    procedure SetConnected(newValue:boolean; UsbDevHandle: Pusb_dev_handle);
  public
    constructor Create;
    destructor Destroy;override;
    
    //after you have assigned the various event handlers you shoul call this
    //method to start the communication with the phone
    procedure Start;
    
    { Methods to display a the name of a contact on the phone.
      These commands are accepted during a short while after the various
      keypress events. DisplayCall and DisplayRedial have the same effect,
      I wrote both for completeness.}
      
    //this one only displays a name, it's used for the "contacts" events,
    //since in that mode it's not possible to see the details of a call
    procedure DisplayContact(s:string);
    
    //this one displays the name but can also display the date/time of
    //the call and the current state of the contact, so it's used for "calls"
    //events.
    procedure DisplayCall(s:string;online:boolean;dt:TDateTime);
    
    //this one does exactly the same as the previous one but it's used for
    //the "redial" events, though they both work in each other context
    procedure DisplayRedial(s:string;online:boolean;dt:TDateTime);
    
    //this one displays the id of the contact after pressing the "Ok" key.
    //It doesn't overwrite the name buffer, so a subsequent "Ok" won't send
    //a message to the pc, and a "Dial" will use the name, not the id
    procedure DisplayId(id:string);

    //Rings the phone and displays the callerid (limited to 20 characters)
    procedure Ring(callerid:string);
    
    //Stops ringing
    procedure StopRing;
    
    //This method will only display the callerid without ringing the phone
    procedure SilentRing(callerid:string);
    
    //This method will hangup the phone during a call, to be used when the
    //call is terminated by the other party or if you want to terminate
    //a call request coming from the phone (i.e. after an OnDialNumber or
    //OnDialName event)
    procedure Hangup;
    
    //Changes the status displayed on the phone
    property ButlerStatus:TButlerStatus read FButlerStatus write SetButlerStatus;
    
    //Activates the calls icon
    property CallsStatus:TCallsStatus read FCallsStatus write SetCallsStatus;
    
    //Notification that the phone has been found
    property OnConnected:TNotifyEvent read FOnConnected write FOnConnected;
    
    //Notification that the phone has been disconnected
    property OnDisconnected:TNotifyEvent read FOnDisconnected write FOnDisconnected;

    //Notification that a number has been dialled on the phone
    //(i.e. number dialled and button pressed)
    property OnDialNumber:TNotifyDial read FOnDialNumber write FOnDialNumber;
    
    //Notification that a contact (coming from a previous Display method call)
    //should be called
    property OnDialName:TNotifyDial read FOnDialName write FOnDialName;
    
    //Notification that the off hook button has been pressed while ringing
    //(i.e. the incoming call should be accepted)
    property OnAcceptCall:TNotifyEvent read FOnAcceptCall write FOnAcceptCall;
    
    //Notification that the handset has hang up (i.e. the current call should be
    //terminated)
    property OnHangup:TNotifyEvent read FOnHangup write FOnHangup;
    
    //A digit has been dialled during a call
    property OnDigit:TNotifyDial read FOnDigit write FOnDigit;
    
    //Notification that the "contacts" button has been pressed
    //(the program should use the DisplayContact method as a reaction to this
    //event)
    property OnContactsKey:TNotifyKey read FOnContactsKey write FOnContactsKey;
    
    //Notification that the "calls" button has been pressed
    //(the program should use the DisplayCall method as a reaction to this
    //event)
    property OnCallsKey:TNotifyKey read FOnCallsKey write FOnCallsKey;
    
    //Notification that the "redial" button has been pressed
    //(the program should use the DisplayRedial method as a reaction to this
    //event)
    property OnRedialKey:TNotifyKey read FOnRedialKey write FOnRedialKey;

    //The id of the name currently on screen is requested
    //(the program should use the DisplayId method as a reaction to this
    //event)
    property OnShowIdKey:TNotifyDial read FOnShowIdKey write FOnShowIdKey;

    //returns true if the phone is connected, false otherwise
    property Connected:boolean read FConnected;
    
    //if a message received from the phone has a wrong checksum
    //ignore the checksum instead of discarding the message
    property IgnoreChecksumErrors:boolean read FIgnoreChecksumErrors write FIgnoreChecksumErrors;
  end;

implementation

{

The typical message to/from the phone is
F0 nn dd dd dd ... cs, where nn is the length of the message (including
the checksum) dd are the data bytes and cs is a checksum (xor of
the data bytes).
The first byte of the message is an command/status code, the rest are
parameters. Codes from 00 to 7F are status codes from the phone, while 80..FF
are commands to the phone.

}

function checksum(s:string):char;
var c,i:integer;
begin
  c:=0;
  for i:=1 to length(s) do c:=c xor ord(s[i]);
  result:=chr(c)
end;

{ TButlerPhone }

constructor TButlerPhone.Create;
begin
    FLastCommand:=' ';
    FConnected:=false;
end;

destructor TButlerPhone.Destroy;
begin
  Log(LOGLEVEL_INFO,'Destroying TButlerPhone');
  if Fconnected then
  begin
    Log(LOGLEVEL_INFO,'Connected: terminating status thread');
    //at this point we're not interested anymore in status replies,
    //stop the thread early so it won't try to send back confirmations
    FStatusThread.Terminate;
    //it may be stuck in Synchronize
    CheckSynchronize();
    //now send the last messages
    Log(LOGLEVEL_INFO,'Stop ringing');
    StopRing;
    //in case the ringer thread is stuck too in Synchronize
    CheckSynchronize();
    Log(LOGLEVEL_INFO,'Hangup');
    HangUp;
    Log(LOGLEVEL_INFO,'Clear display');
    ButlerStatus:=BSClearDisplay;
    CallsStatus:=CSNoCalls;
    //wait for the send thread to empty the fifo before stopping it
    Log(LOGLEVEL_INFO,'Waiting for send fifo empty');
    while not FSendFifo.IsEmpty do sleep(100);
    Log(LOGLEVEL_INFO,'Terminating send thread');
    FSendThread.Terminate;
    FSendThread.Free;
    Log(LOGLEVEL_INFO,'Destroying send fifo');
    FSendFifo.Free;
    //if the status thread hasn't terminated yet, it will now
    Log(LOGLEVEL_INFO,'Reset and close device');
    usb_reset(FUsbDevHandle);
    usb_release_interface(FUsbDevHandle,0);
    usb_close(FUsbDevHandle);
  end else
  begin
    Log(LOGLEVEL_INFO,'Not connected: terminating connect thread');
    FConnectThread.Terminate;
  end;
  inherited Destroy;
end;

//looks for the phone
procedure TButlerPhone.Start;
begin
  if not FConnected and (FConnectThread=nil) then
    FConnectThread:=TConnectThread.create(self);
end;

procedure TButlerPhone.SetConnected(newValue: boolean; UsbDevHandle:Pusb_dev_handle);
begin
  if newValue=FConnected then exit;
  FConnected:=newValue;
  if FConnected then
  begin
    Log(LOGLEVEL_INFO,'connected');
    //start the various threads and update the display
    FConnectThread:=nil;
    FUsbDevHandle:=UsbDevHandle;
    FStatusThread:=TStatusThread.create(self,UsbDevHandle);
    FSendFifo:=TSendFifo.create;
    FSendThread:=TSendThread.create(UsbDevHandle,FSendFifo);
    SendStatus;
    if assigned(FOnConnected) then FOnConnected(self);
  end else
  begin
    Log(LOGLEVEL_INFO,'disconnected');
    //stop everything and start waiting for a new connection
    if FRingerThread<>nil then
    begin
      FRingerThread.Terminate;
      FRingerThread:=nil;
    end;
    FStatusThread.Terminate;
    FSendThread.Terminate;
    FSendFifo.Free;
    usb_reset(FUsbDevHandle);
    usb_release_interface(FUsbDevHandle,0);
    usb_close(FUsbDevHandle);
    FConnectThread:=TConnectThread.create(self);
    if assigned(FOnDisconnected) then FOnDisconnected(self);
  end;
end;

// This method is called from TStatusThread for each status message
// Not all status codes are known
//
procedure TButlerPhone.receive(s: string);
var l:integer;

    {
    Unified key processing:
       XX FF FF 00 : XX open
       XX 00 00 00 : XX down
       XX 00 00 01 : XX up
    }
    
    procedure ManageKey(handler:TNotifyKey);
    var
        LocKey:TContactKey;
    begin
      if (length(s)<4) or not assigned(handler) then exit;
      if s[2]=chr($ff) then LocKey:=CKFirst else
      if s[4]=chr($00) then LocKey:=CKDown  else
                            LocKey:=CKUp;
      Handler(self,LocKey);
    end;
    
begin
  LogHex(LOGLEVEL_DEBUG,'Received:',s);
  l:=length(s);
  if l<1 then exit;
  case s[1] of
    //------- 7E : are you there?
    chr($7e):send(chr($7e)); //yes I am
    
    //------- 23 : parameters of the phone (reply to 85 01)
    //             Ignored since I don't know what they mean
    chr($23):;
    
    //------- 24 : you didn't reply, I'm blanking
    chr($24): SendStatus;
     
    //------- 50 : contacts keys
    chr($50): ManageKey(FOnContactsKey);
    
    //------- 53 : calls keys
    chr($53): ManageKey(FOnCallsKey);
    
    //------- 55 : calls exit
    chr($55): 
      begin
        if assigned(FOnCallsKey) then FOnCallsKey(self,CKExit);
        SendStatus;
      end;  
      
    //------- 54 : redial keys
    chr($54): ManageKey(FOnRedialKey);
      
    //------- 51 n 01 CC CC ... ok (showid) button
    chr($51):
      begin
        if l<4 then exit;
        if assigned(FOnShowIdKey) then FOnShowIdKey(self,copy(s,4,ord(s[1])));
      end;
                
    //------- 30 01 x DD DD ... dialled digits
    //------- 30 00 x CC CC ... dialled characters (data on display)
    chr($30):
      begin
        if l<4 then exit;
        if (s[2]<>chr($01)) and (s[2]<>chr($00)) then exit;
        send(s[1]);
        send(chr($91)+chr($01));
        if s[2]=chr($01) then
        begin
          if assigned(FOnDialNumber)  then FOnDialNumber(self, DecodeNumber(ord(s[3]),copy(s,4,l-3)));
        end else
        begin
          if assigned(FOnDialName) then FOnDialName(self, copy(s,5,ord(s[3])));
        end;
      end;
      
    //------- 35 : silent ringing
    chr($35): send(chr($35)); //confirm and ignore it
    
    //------- 36 : accept call (after silent ringing)
    //------- 31 : accept call (after ringing)
    chr($36),chr($31):
      begin
        StopRing;
        send(s[1]);
        send(chr($91)+chr($01));
        if assigned(FOnAcceptCall) then FOnAcceptCall(self);
      end;
      
    //------- 32 : hang up
    chr($32):
      begin
        send(chr($32));
        SendStatus; //to avoid auto blinking of the calls icon
        if assigned(FOnHangup) then FOnHangup(self);
      end;
      
    //------- 34 01 DD : dialled digit (during a call)
    chr($34):
      begin
        if l<3 then exit;
        if assigned(FOnDigit) then FOnDigit(self,DecodeNumber(1,s[3]));
      end;

    else
      begin
        //FIXME check if reply corresponds to command or simply ignore it?
      end;
  end; //case s[1] of
end;

procedure TButlerPhone.Command(s: string);
begin
  if length(s)<1 then exit;
  FLastCommand:=s[1];
  Send(s);
end;

procedure TButlerPhone.Send(s: string);
begin
  if not FConnected then exit;
  LogHex(LOGLEVEL_DEBUG,'Message:',s);
  FSendFifo.Push(s);
end;

procedure TButlerPhone.SendStatus;
begin
  Command(chr($84)+chr(ord(FButlerStatus)));
  Command(chr($8c)+chr(ord(FButlerStatus))+chr(ord(FCallsStatus))+chr(0)+chr(0)+chr(0)+chr(0)+chr(0)+chr(0)+chr(0));
end;

//each byte has two digit, the 1st in the right nibble and the 2nd in the left
function TButlerPhone.DecodeNumber(digits: integer; code: string): string;
const codes:array[0..15] of char='0123456789*# P  ';
var i,n:integer;
begin
  if length(code)<(digits+1) div 2 then
  begin
    result:='';
    exit;
  end;
  setlength(result,digits);
  for i:=0 to digits-1 do
  begin
    n:=i div 2;
    if n*2=i then result[i+1]:=codes[ord(code[n+1]) and $0f]
             else result[i+1]:=codes[(ord(code[n+1]) div 16)]
  end;
end;

procedure TButlerPhone.SetButlerStatus(newValue: TButlerStatus);
begin
  if newValue<>FButlerStatus then
  begin
    FButlerStatus:=newValue;
    SendStatus;
  end;
end;

procedure TButlerPhone.SetCallsStatus(newValue: TCallsStatus);
begin
  if newValue<>FCallsStatus then
  begin
    FCallsStatus:=newValue;
    SendStatus;
  end;
end;

procedure TButlerPhone.DisplayContact(s: string);
begin
  if length(s)>20 then setlength(s,20);
  Command(chr($86)+chr($01)+chr($00)+chr(length(s))+chr($01)+s);
end;

procedure TButlerPhone.DisplayId(id: string);
begin
  if length(id)>20 then setlength(id,20);
  Command(chr($87)+chr(length(id))+id);
end;

procedure TButlerPhone.DisplayCall(s: string; online: boolean; dt: TDateTime);
var
  chonline:char;
  year,month,day,hour,minute,second,ms:word;
begin
  DecodeDate(dt,year,month,day);
  DecodeTime(dt,hour,minute,second,ms);
  if online then chonline:=chr($10) else chonline:=chr($01);
  if length(s)>20 then setlength(s,20);
  Command(chr($89)+chr($00)+chr($00)+chonline+chr(year)+chr(month)+chr(day)+
          chr(hour)+chr(minute)+chr(second)+chr(length(s))+chr($01)+s);
end;

procedure TButlerPhone.DisplayRedial(s: string; online: boolean; dt: TDateTime);
var
  chonline:char;
  year,month,day,hour,minute,second,ms:word;
begin
  DecodeDate(dt,year,month,day);
  DecodeTime(dt,hour,minute,second,ms);
  if online then chonline:=chr($10) else chonline:=chr($00);
  if length(s)>20 then setlength(s,20);
  Command(chr($8a)+chr($00)+chr($00)+chonline+chr(year)+chr(month)+chr(day)+
          chr(hour)+chr(minute)+chr(second)+chr(length(s))+chr($01)+s);

end;

procedure TButlerPhone.Ring(callerid: string);
begin
  if length(callerid)>20 then setlength(callerid,20);
  Command(chr($80)+chr($01));
  Command(chr($81)+chr(length(callerid))+chr($00)+callerid+chr($01));
  if FRingerThread=nil then FRingerThread:=TRingerThread.Create(self);
end;

procedure TButlerPhone.StopRing;
begin
  if FRingerThread<>nil then
  begin
    FRingerThread.Terminate;
    FRingerThread:=nil;
  end;
  Command(chr($80)+chr($00));
  SendStatus; //avoid auto icon blinking
end;

procedure TButlerPhone.SilentRing(callerid: string);
begin
  Command(chr($90)+chr(length(callerid))+chr($00)+callerid+chr($01));
end;

procedure TButlerPhone.Hangup;
begin
  Command(chr($82)+chr($82));
  SendStatus; //avoid auto icon blinking
end;

{ TRingerThread }

constructor TRingerThread.create(AOwner: TButlerPhone);
begin
  FOwner:=AOwner;
  FreeOnTerminate:=True;
  inherited create(false);
end;

procedure TRingerThread.SyncProc;
begin
  FOwner.Command(chr($80)+chr($01));
end;

procedure TRingerThread.execute;
begin
  while not terminated do
  try
    sleep(3000);
    if terminated then break;
    Synchronize(@syncproc);
  except
    on e:exception do Log(LOGLEVEL_ERROR,'TRingerThread exception '+e.message);
  end;
  //terminated
end;

{ TStatusThread }

constructor TStatusThread.create(AOwner: TButlerPhone; UsbDevHandle:Pusb_dev_handle);
begin
  FOwner:=AOwner;
  FUsbDevHandle:=UsbDevHandle;
  FreeOnTerminate:=true;
  inherited create(false);
end;

procedure TStatusThread.Receive;
begin
  FOwner.Receive(FData);
end;

procedure TStatusThread.Error;
begin
  FOwner.SetConnected(false,nil);
end;

procedure TStatusThread.execute;
var buffer:array[1..$80] of char;
    b2:array[1..2] of char;
    r,i,r2,l:integer;
    messageok:boolean;
    goodchecksum:boolean;
begin
  {
  
  Status from the phone is read using a control message.
  The status comes in 2 messages: the first one with the data and the
  second one with the length of the data received in the first one (this is
  the reason they don't release specification: the protocol is stupid and
  they're embarassed to show it in public)
  But it's not over: a single message can contain more than one status message
  (in case the pc did not poll fast enough).
  
  }
  While not Terminated do
  try
    //read the status message(s)
    r:=usb_control_msg(FUsbDevHandle, $c0, $04, $5002, $0000, @buffer, $80, 0);
    if terminated then break;
    if r>=0 then
    begin
      //read the length of the previous message (braindead)
      r2:=usb_control_msg(FUsbDevHandle, $c0, $04, $500d, $0000, @b2, 2, 100);
      if terminated then break;
      if r2=2 then
      begin
        r2:=ord(b2[1])*256+ord(b2[2]);
        if r2=r then
        begin
          messageok:=true;
        end else
        begin
          messageok:=false;
          Log(LOGLEVEL_DEBUG,'wrong length in second message, expected '+IntToStr(r)+' received '+IntToStr(r2));
        end;
      end else
      begin
        messageok:=false;
        Log(LOGLEVEL_DEBUG,'second message not received');
      end;
      
      if messageok then
      begin
        i:=1;
        //check all possible status messages
        while i<r do
        begin
          if buffer[i]<>chr($f0) then
          begin
            Log(LOGLEVEL_DEBUG,'status doesn''t start with F0');
            break;
          end;
          l:=ord(buffer[i+1]);
          FData:=copy(buffer,i+2,l-1);
          goodchecksum:=buffer[i+l+1]=checksum(FData);
          if not goodchecksum then
          begin
            LogHex(LOGLEVEL_DEBUG,'wrong checksum '+IntToHex(ord(buffer[i+l+1]),2)+' for',FData);
            if FOwner.IgnoreChecksumErrors then Log(LOGLEVEL_DEBUG,'(ignored)');
          end;
          if goodchecksum or FOwner.IgnoreChecksumErrors then
          begin
           if not terminated then synchronize(@Receive)
          end;
          i:=i+l+1;
        end;
      end;
    end else
    begin
      Log(LOGLEVEL_INFO,'Status read failed');
      if not terminated then Synchronize(@error);
      break;
    end;
  except
    on e:exception do Log(LOGLEVEL_ERROR,'TStatusThread exception '+e.message);
  end;  
  //terminated
  Log(LOGLEVEL_INFO,'Status thread terminated');
end;

{ TSendThread }

constructor TSendThread.create(UsbDevHandle:Pusb_dev_handle;SendFifo:TSendFifo);
begin
  FUsbDevHandle:=UsbDevHandle;
  FSendFifo:=SendFifo;
  FWaiter:=TEventObject.Create(nil,false,false,'');
  FSendFifo.Waiter:=FWaiter;
  FreeOnTerminate:=false;
  FTerminated:=false;
  inherited create(false);
end;

procedure TSendThread.terminate;
begin
  inherited terminate;
  FWaiter.SetEvent;
  while not FTerminated do sleep(100);
end;

procedure TSendThread.execute;
var buf:array[1..8] of char;
    m:string;
    i,l,i2,len,r:integer;
begin
  while not terminated do
  try
    while true do
    begin
       if FSendFifo.IsEmpty then FWaiter.WaitFor($FFFFFFFF);
       if terminated then break;
       if FSendFifo.Pop(m) then break;
    end;
    if terminated then break;
    //send data
    m:=m+checksum(m);
    l:=length(m);
    m:=chr($f0)+chr(l)+m;
    l:=length(m);
    i2:=1;
    {
    Now that we have a standard message (F0 nn dd dd ... cs) we have to send it.
    Messages are sent with an interrupt write, but in each call only and alway
    8 bytes are sent, the first one is '8x' (where x=1..7 is the number of
    data bytes), the rest are the data bytes (padded with 0s if there are less
    than 7)
    }
    while l>0 do
    begin
      len:=l;
      if len>7 then len:=7;
      l:=l-len;
      buf[1]:=chr($80+len);
      for i:=2 to len+1 do
      begin
        buf[i]:=m[i2];
        i2:=i2+1;
      end;
      if len<7 then for i:=len+2 to 8 do buf[i]:=chr(0);
      LogHex(LOGLEVEL_DEBUG,'Sending:',buf);
      r:=usb_interrupt_write(FUsbDevHandle, 3, @buf, 8, 200);
      if r<>8 then
      begin
        Log(LOGLEVEL_DEBUG,'Sending failed: '+usb_strerror);
      end;
      //we have to sleep here otherwise the phone can't keep up
      sleep(100);
    end;
  except
    on e:exception do Log(LOGLEVEL_ERROR,'TSendThread exception '+e.message);
  end;  
  //terminated
  FWaiter.Free;
  Log(LOGLEVEL_INFO,'Send thread terminated');
  FTerminated:=true;
end;

{ TSendFifo }

constructor TSendFifo.Create;
begin
  FWaiter:=nil;
  FFifo:=TStringList.Create;
  FAccess:=TCriticalSection.Create;
end;

destructor TSendFifo.Destroy;
begin
  FFifo.Free;
  FAccess.Free;
  inherited Destroy;
end;

procedure TSendFifo.Push(s: string);
begin
  FAccess.Acquire;
  FFifo.Add(s);
  if (FWaiter<>nil) and (FFifo.Count=1) then FWaiter.SetEvent;
  FAccess.Release;
end;

function TSendFifo.Pop(var s: string): boolean;
begin
  FAccess.Acquire;
  result:=FFifo.Count>0;
  if result then
  begin
    s:=FFifo[0];
    FFifo.delete(0);
  end;
  FAccess.Release;
end;

function TSendFifo.IsEmpty: boolean;
begin
  FAccess.Acquire;
  result:=FFifo.Count=0;
  FAccess.Release;
end;

{ TConnectThread }

constructor TConnectThread.create(AOwner: TButlerPhone);
begin
  FOwner:=AOwner;
  inherited create(false);
  FreeOnTerminate:=true;
end;

procedure TConnectThread.SyncProc;
begin
  FOwner.SetConnected(True,FUsbDevHandle);
end;

function TConnectThread.FindDevice:pusbdevice;
const
  vid = $E6A;
  pid = $6002;
var
  bus : pusb_bus;
  dev : pusbdevice;
begin
  usb_init();
  usb_find_busses();
  usb_find_devices();
  bus:=usb_get_busses;
  while bus<>nil do
  begin
    dev:=bus^.devices;
    while dev<>nil do
    begin
      if (dev^.descriptor.idVendor=vid) and (dev^.descriptor.idProduct=pid) then
      begin
        result:=dev;
        exit;
      end;
      dev:=dev^.next;
    end;
    bus:=bus^.next;
  end;
  result:=nil;
end;

function TConnectThread.InitDevice:boolean;
var buf:array[1..2] of char;
begin
  usb_reset(FUsbDevHandle);
  if usb_control_msg(FUsbDevHandle, $c0, $04, $5004, $0f09, @buf, 2, 100)=2 then
  begin
    buf[1]:=chr($90);
    if usb_control_msg(FUsbDevHandle, $40, $0c, $5003, $00f0, @buf, 1, 100)=1 then
    begin
      buf[1]:=chr($08);
      if usb_control_msg(FUsbDevHandle, $40, $0c, $5003, $ffff, @buf, 1, 100)=1 then
      begin
        result:=true;
      end else
      begin
        Log(LOGLEVEL_ERROR,'Init device failed send 08 : '+usb_strerror);
        result:=false;
      end
    end else
    begin
      Log(LOGLEVEL_ERROR,'Init device failed send 90 : '+usb_strerror);
      result:=false;
    end
  end else
  begin
    Log(LOGLEVEL_ERROR,'Init device failed initial read : '+usb_strerror);
    result:=false;
  end;
end;

procedure TConnectThread.Execute;
var
  dev : pusbdevice;
begin
  Log(LOGLEVEL_INFO,'Waiting for device connected');
  while not terminated do
  try
    dev:=FindDevice;
    if dev<>nil then
    begin
       Log(LOGLEVEL_INFO,'Found device at '+dev^.bus^.dirname+'/'+dev^.filename+' opening it...');
       FusbDevHandle:=usb_open(dev);
       if FUsbDevHandle<>nil then
       begin
         Log(LOGLEVEL_INFO,'Claiming interface...');
         if (usb_claim_interface(FUsbDevHandle,0)=0) then
         begin
           Log(LOGLEVEL_INFO,'Initializing device..');
           if InitDevice then
           begin
             Log(LOGLEVEL_INFO,'Initialized');
             Synchronize(@SyncProc);
             exit;
           end else
           begin
             usb_close(FUsbDevHandle);
           end;
         end else
         begin
           Log(LOGLEVEL_INFO,'Claim failed: '+usb_strerror);
           usb_close(FUsbDevHandle);
         end;
       end else Log(LOGLEVEL_INFO,'Open failed: '+usb_strerror);
    end;
    sleep(1000);
  except
    on e:exception do LOG(LOGLEVEL_ERROR,'TConnectThread exception '+e.message);
  end;  
end;

end.

